import { randomBytes } from 'crypto';

import { EventEmitter } from 'events';
import { IPlugin } from '../plugin';
import { IMessage, IMessageTarget } from './message';
import { Formatter } from './message-format';

/**
 * The base class for a protocol handler.
 */
export class Protocol extends EventEmitter {
  public format: Formatter = new Formatter(false, false);

  public id = randomBytes(4).toString('hex');

  // override this!
  public type = 'GenericProtocol';

  public running = false;

  // This should be set to true when the protocol was stopped for any reason
  // at any time.
  public stopped = false;

  // This should be set to true when the protocol fails for any reason
  public failed = false;

  // This should identify the bot itself
  public me: IMessageTarget = {
    id: '',
    name: '',
  };

  constructor(public plugin: IPlugin, public config: any) {
    super();
    this.passEvents();
  }

  public get name(): string {
    return this.config.name;
  }

  public get status(): boolean {
    return this.running;
  }

  public start(...args: any[]): void {
    this.running = true;
    this.emit('running');
  }

  public stop(force = false): void {
    if (!this.running) {
      return;
    }

    this.running = false;
    this.stopped = true;

    if (force) {
      this.failed = true;
    }

    this.emit('stopped');
  }

  public resolve(message: IMessage, ...data: any[]): void {
    throw new Error('Method not implemented.');
  }

  public async sendTo(target: string, ...data: any[]): Promise<boolean> {
    throw new Error('Method not implemented.');
  }

  public get fullName(): string {
    return this.plugin.manifest.name + '/' + this.name;
  }

  protected passEvents(): void {
    this.on('stop', (force) => this.stop(force));
    this.on('start', (...args: any[]) => {
      // Prevent restarting from here, it's unsafe
      if (this.status || (!this.status && this.stopped)) {
        return;
      }
      this.start(...args);
    });
  }
}
